@@ -65,7 +65,7 @@ var Fetch = {
   initFetchWorker: function() {
     var stackSize = 128*1024;
     var stack = allocate(stackSize>>2, "i32*", ALLOC_DYNAMIC);
-    Fetch.worker.postMessage({cmd: 'init', TOTAL_MEMORY: TOTAL_MEMORY, DYNAMICTOP_PTR: DYNAMICTOP_PTR, STACKTOP: stack, STACK_MAX: stack + stackSize, queuePtr: _fetch_work_queue, buffer: HEAPU8.buffer}, [HEAPU8.buffer]);
+    Fetch.worker.postMessage({cmd: 'init', TOTAL_MEMORY: TOTAL_MEMORY, DYNAMICTOP_PTR: DYNAMICTOP_PTR, STACKTOP: stack, STACK_MAX: stack + stackSize, queuePtr: _fetch_work_queue, buffer: HEAPU8.buffer});
   },
 
   staticInit: function() {
@@ -83,7 +83,7 @@ var Fetch = {
 
 #if USE_PTHREADS
       if (isMainThread) {
-        Fetch.initFetchWorker();
+        if (typeof SharedArrayBuffer !== 'undefined') Fetch.initFetchWorker();
         removeRunDependency('library_fetch_init');
       }
 #else
@@ -98,7 +98,7 @@ var Fetch = {
 
 #if USE_PTHREADS
       if (isMainThread) {
-        Fetch.initFetchWorker();
+        if (typeof SharedArrayBuffer !== 'undefined') Fetch.initFetchWorker();
         removeRunDependency('library_fetch_init');
       }
 #endif
@@ -478,7 +478,7 @@ function emscripten_start_fetch(fetch, successcb, errorcb, progresscb) {
 #if FETCH_DEBUG
     console.log('fetch: operation success. e: ' + e);
 #endif
-    if (onsuccess && Runtime.dynCall) Runtime.dynCall('vi', onsuccess, [fetch]);
+    if (onsuccess && Runtime.dynCall) Module['dynCall_vi'](onsuccess, fetch);
     else if (successcb) successcb(fetch);
   };
 
@@ -497,12 +497,12 @@ function emscripten_start_fetch(fetch, successcb, errorcb, progresscb) {
 #endif
     };
     __emscripten_fetch_cache_data(Fetch.dbInstance, fetch, xhr.response, storeSuccess, storeError);
-    if (onsuccess && Runtime.dynCall) Runtime.dynCall('vi', onsuccess, [fetch]);
+    if (onsuccess && Runtime.dynCall) Module['dynCall_vi'](onsuccess, fetch);
     else if (successcb) successcb(fetch);
   };
 
   var reportProgress = function(fetch, xhr, e) {
-    if (onprogress && Runtime.dynCall) Runtime.dynCall('vi', onprogress, [fetch]);      
+    if (onprogress && Runtime.dynCall) Module['dynCall_vi'](onprogress, fetch);
     else if (progresscb) progresscb(fetch);
   };
 
@@ -510,7 +510,7 @@ function emscripten_start_fetch(fetch, successcb, errorcb, progresscb) {
 #if FETCH_DEBUG
     console.error('fetch: operation failed: ' + e);
 #endif
-    if (onerror && Runtime.dynCall) Runtime.dynCall('vi', onerror, [fetch]);
+    if (onerror && Runtime.dynCall) Module['dynCall_vi'](onerror, fetch);
     else if (errorcb) errorcb(fetch);
   };
 
