@@ -18,6 +18,7 @@
 #include <sys/mman.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
+#include "../internal/libc.h"
 #include "../internal/pthread_impl.h"
 #include <assert.h>
 
@@ -31,6 +32,8 @@ char *gets(char *);
 // Extra pthread_attr_t field:
 #define _a_transferredcanvases __u.__s[9]
 
+void __pthread_testcancel();
+
 int emscripten_pthread_attr_gettransferredcanvases(const pthread_attr_t *a, const char **str)
 {
 	*str = (const char *)a->_a_transferredcanvases;
@@ -54,7 +57,7 @@ static void inline __pthread_mutex_locked(pthread_mutex_t *mutex)
 	assert(mutex);
 	assert(mutex->_m_lock == 0);
 	mutex->_m_lock = pthread_self()->tid;
-	if (_pthread_getcanceltype() == PTHREAD_CANCEL_ASYNCHRONOUS) pthread_testcancel();
+	if (_pthread_getcanceltype() == PTHREAD_CANCEL_ASYNCHRONOUS) __pthread_testcancel();
 }
 
 double _pthread_msecs_until(const struct timespec *restrict at)
@@ -93,7 +96,7 @@ int pthread_mutex_lock(pthread_mutex_t *mutex)
 				double msecs = INFINITY;
 				if (threadCancelType == PTHREAD_CANCEL_ASYNCHRONOUS) {
 					// Sleep in small slices so that we can test cancellation to honor PTHREAD_CANCEL_ASYNCHRONOUS.
-					pthread_testcancel();
+					__pthread_testcancel();
 					msecs = 100;
 				}
 				emscripten_futex_wait(&mutex->_m_addr, 2, msecs);
@@ -174,7 +177,7 @@ int pthread_mutex_timedlock(pthread_mutex_t *restrict mutex, const struct timesp
 				// Sleep in small slices if thread type is PTHREAD_CANCEL_ASYNCHRONOUS
 				// so that we can honor PTHREAD_CANCEL_ASYNCHRONOUS requests.
 				if (threadCancelType == PTHREAD_CANCEL_ASYNCHRONOUS) {
-					pthread_testcancel();
+					__pthread_testcancel();
 					if (msecs > 100) msecs = 100;
 				}
 				int ret = emscripten_futex_wait(&mutex->_m_addr, 2, msecs);
@@ -227,7 +230,7 @@ int _pthread_isduecanceled(struct pthread *pthread_ptr)
 	return pthread_ptr->threadStatus == 2/*canceled*/;
 }
 
-void pthread_testcancel()
+void __pthread_testcancel()
 {
 	struct pthread *self = pthread_self();
 	if (self->canceldisable) return;
@@ -247,17 +250,17 @@ int pthread_getattr_np(pthread_t t, pthread_attr_t *a)
 
 static uint32_t dummyZeroAddress = 0;
 
-int usleep(unsigned usec)
+static void do_sleep(double msecs)
 {
 	int is_main_thread = emscripten_is_main_runtime_thread();
 	double now = emscripten_get_now();
-	double target = now + usec * 1e-3;
+	double target = now + msecs;
 #ifdef __EMSCRIPTEN__
 	emscripten_conditional_set_current_thread_status(EM_THREAD_STATUS_RUNNING, EM_THREAD_STATUS_SLEEPING);
 #endif
 	while(now < target) {
 		if (is_main_thread) emscripten_main_thread_process_queued_calls(); // Assist other threads by executing proxied operations that are effectively singlethreaded.
-		pthread_testcancel(); // pthreads spec: usleep is a cancellation point, so it must test if this thread is cancelled during the sleep.
+		__pthread_testcancel(); // pthreads spec: usleep is a cancellation point, so it must test if this thread is cancelled during the sleep.
 		now = emscripten_get_now();
 		double msecsToSleep = target - now;
 		if (msecsToSleep > 1.0) {
@@ -268,13 +271,43 @@ int usleep(unsigned usec)
 	}
 #ifdef __EMSCRIPTEN__
 	emscripten_conditional_set_current_thread_status(EM_THREAD_STATUS_SLEEPING, EM_THREAD_STATUS_RUNNING);
-#endif
+#endif	
+}
+
+int nanosleep(const struct timespec *req, struct timespec *rem)
+{
+	if (!req || req->tv_nsec < 0 || req->tv_nsec > 999999999L || req->tv_sec < 0) return EINVAL;
+	do_sleep(req->tv_sec * 1000.0 + req->tv_nsec / 1e6);
 	return 0;
 }
 
+int usleep(unsigned usec)
+{
+	do_sleep(usec / 1e3);
+	return 0;
+}
+
+// Allocator and deallocator for em_queued_call objects.
+static em_queued_call *em_queued_call_malloc()
+{
+	em_queued_call *call = (em_queued_call*)malloc(sizeof(em_queued_call));
+	call->operationDone = 0;
+	call->functionPtr = 0;
+	return call;
+}
+static void em_queued_call_free(em_queued_call *call)
+{
+	free(call);
+}
+
+void emscripten_async_waitable_close(em_queued_call *call)
+{
+	em_queued_call_free(call);
+}
+
 static void _do_call(em_queued_call *q)
 {
-	switch(q->function)
+	switch(q->functionEnum)
 	{
 		case EM_PROXIED_UTIME: q->returnValue.i = utime(q->args[0].cp, (struct utimbuf*)q->args[1].vp); break;
 		case EM_PROXIED_UTIMES: q->returnValue.i = utimes(q->args[0].cp, (struct timeval*)q->args[1].vp); break;
@@ -291,46 +324,111 @@ static void _do_call(em_queued_call *q)
 		case EM_PROXIED_TZSET: tzset(); break;
 		case EM_PROXIED_PTHREAD_CREATE: q->returnValue.i = pthread_create(q->args[0].vp, q->args[1].vp, q->args[2].vp, q->args[3].vp); break;
 		case EM_PROXIED_SYSCALL: q->returnValue.i = emscripten_syscall(q->args[0].i, q->args[1].vp); break;
+		case EM_FUNC_SIG_V: ((em_func_v)q->functionPtr)(); break;
+		case EM_FUNC_SIG_VI: ((em_func_vi)q->functionPtr)(q->args[0].i); break;
+		case EM_FUNC_SIG_VII: ((em_func_vii)q->functionPtr)(q->args[0].i, q->args[1].i); break;
+		case EM_FUNC_SIG_VIII: ((em_func_viii)q->functionPtr)(q->args[0].i, q->args[1].i, q->args[2].i); break;
+		case EM_FUNC_SIG_I: q->returnValue.i = ((em_func_i)q->functionPtr)(); break;
+		case EM_FUNC_SIG_II: q->returnValue.i = ((em_func_ii)q->functionPtr)(q->args[0].i); break;
+		case EM_FUNC_SIG_III: q->returnValue.i = ((em_func_iii)q->functionPtr)(q->args[0].i, q->args[1].i); break;
+		case EM_FUNC_SIG_IIII: q->returnValue.i = ((em_func_iiii)q->functionPtr)(q->args[0].i, q->args[1].i, q->args[2].i); break;
 		default: assert(0 && "Invalid Emscripten pthread _do_call opcode!");
 	}
-	q->operationDone = 1;
-	emscripten_futex_wake(&q->operationDone, INT_MAX);
+
+	// If the caller is detached from this operation, it is the main thread's responsibility to free up the call object.
+	if (q->calleeDelete) {
+		emscripten_async_waitable_close(q);
+		// No need to wake a listener, nothing is listening to this since the call object is detached.
+	} else {
+		// The caller owns this call object, it is listening to it and will free it up.
+		q->operationDone = 1;
+		emscripten_futex_wake(&q->operationDone, INT_MAX);
+	}
 }
 
 #define CALL_QUEUE_SIZE 128
 static em_queued_call **call_queue = 0;
-static int call_queue_length = 0; // Shared data synchronized by call_queue_lock.
+static volatile int call_queue_head = 0; // Shared data synchronized by call_queue_lock.
+static volatile int call_queue_tail = 0;
 static pthread_mutex_t call_queue_lock = PTHREAD_MUTEX_INITIALIZER;
 
-void EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread(em_queued_call *call)
+EMSCRIPTEN_RESULT emscripten_wait_for_call_v(em_queued_call *call, double timeoutMSecs)
+{
+	int r;
+
+	int done = emscripten_atomic_load_u32(&call->operationDone);
+	if (!done) {
+		double now = emscripten_get_now();
+		double waitEndTime = now + timeoutMSecs;
+		emscripten_set_current_thread_status(EM_THREAD_STATUS_WAITPROXY);
+		while(!done && now < waitEndTime) {
+			r = emscripten_futex_wait(&call->operationDone, 0, waitEndTime - now);
+			done = emscripten_atomic_load_u32(&call->operationDone);
+			now = emscripten_get_now();
+		}
+		emscripten_set_current_thread_status(EM_THREAD_STATUS_RUNNING);
+	}
+	if (done) return EMSCRIPTEN_RESULT_SUCCESS;
+	else return EMSCRIPTEN_RESULT_TIMED_OUT;
+}
+
+EMSCRIPTEN_RESULT emscripten_wait_for_call_i(em_queued_call *call, double timeoutMSecs, int *outResult)
+{
+	EMSCRIPTEN_RESULT res = emscripten_wait_for_call_v(call, timeoutMSecs);
+	if (res == EMSCRIPTEN_RESULT_SUCCESS && outResult) *outResult = call->returnValue.i;
+	return res;
+}
+
+void EMSCRIPTEN_KEEPALIVE emscripten_async_run_in_main_thread(em_queued_call *call)
 {
 	assert(call);
+	// If we are the main Emscripten runtime thread, we can just call the operation directly.
 	if (emscripten_is_main_runtime_thread()) {
 		_do_call(call);
 		return;
 	}
+
+	// Add the operation to the call queue of the main runtime thread.
 	pthread_mutex_lock(&call_queue_lock);
 	if (!call_queue) call_queue = malloc(sizeof(em_queued_call*) * CALL_QUEUE_SIZE); // Shared data synchronized by call_queue_lock.
-	// Note: currently call_queue_length can be at most the number of pthreads that are currently running, so the queue can never get
-	// full. However if/when the queue is extended to be asynchronous for void-returning functions later, this will need to be revised.
-	assert(call_queue_length < CALL_QUEUE_SIZE);
-	call_queue[call_queue_length] = call;
-	++call_queue_length;
-	if (call_queue_length == 1) {
+
+	int head = emscripten_atomic_load_u32((void*)&call_queue_head);
+	int tail = emscripten_atomic_load_u32((void*)&call_queue_tail);
+	int new_tail = (tail + 1) % CALL_QUEUE_SIZE;
+
+	while(new_tail == head) { // Queue is full?
+		pthread_mutex_unlock(&call_queue_lock);
+		emscripten_futex_wait((void*)&call_queue_head, head, INFINITY);
+		pthread_mutex_lock(&call_queue_lock);
+		head = emscripten_atomic_load_u32((void*)&call_queue_head);
+		tail = emscripten_atomic_load_u32((void*)&call_queue_tail);
+		new_tail = (tail + 1) % CALL_QUEUE_SIZE;
+	}
+
+	call_queue[tail] = call;
+
+	// If the call queue was empty, the main runtime thread is likely idle in the browser event loop,
+	// so send a message to it to ensure that it wakes up to start processing the command we have posted.
+	if (head == tail) {
 		EM_ASM(postMessage({ cmd: 'processQueuedMainThreadWork' }));
 	}
+
+	emscripten_atomic_store_u32((void*)&call_queue_tail, new_tail);
+
 	pthread_mutex_unlock(&call_queue_lock);
-	int r;
-	emscripten_set_current_thread_status(EM_THREAD_STATUS_WAITPROXY);
-	do {
-		r = emscripten_futex_wait(&call->operationDone, 0, INFINITY);
-	} while(r != 0 && call->operationDone == 0);
-	emscripten_set_current_thread_status(EM_THREAD_STATUS_RUNNING);
+}
+
+void EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread(em_queued_call *call)
+{
+	emscripten_async_run_in_main_thread(call);
+
+	// Enter to wait for the operation to complete.
+	emscripten_wait_for_call_v(call, INFINITY);
 }
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_0(int function)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.returnValue.vp = 0;
 	emscripten_sync_run_in_main_thread(&q);
 	return q.returnValue.vp;
@@ -338,7 +436,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_0(int function)
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_1(int function, void *arg1)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.returnValue.vp = 0;
 	emscripten_sync_run_in_main_thread(&q);
@@ -347,7 +445,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_1(int function, v
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_2(int function, void *arg1, void *arg2)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.returnValue.vp = 0;
@@ -369,7 +467,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_xprintf_varargs(i
 		va_start(args, format);
 		len = vsnprintf(s, len+1, format, args);
 	}
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = (void*)param0;
 	q.args[1].vp = s;
 	q.returnValue.vp = 0;
@@ -380,7 +478,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_xprintf_varargs(i
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_3(int function, void *arg1, void *arg2, void *arg3)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.args[2].vp = arg3;
@@ -391,7 +489,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_3(int function, v
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_4(int function, void *arg1, void *arg2, void *arg3, void *arg4)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.args[2].vp = arg3;
@@ -403,7 +501,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_4(int function, v
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_5(int function, void *arg1, void *arg2, void *arg3, void *arg4, void *arg5)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.args[2].vp = arg3;
@@ -416,7 +514,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_5(int function, v
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_6(int function, void *arg1, void *arg2, void *arg3, void *arg4, void *arg5, void *arg6)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.args[2].vp = arg3;
@@ -430,7 +528,7 @@ void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_6(int function, v
 
 void * EMSCRIPTEN_KEEPALIVE emscripten_sync_run_in_main_thread_7(int function, void *arg1, void *arg2, void *arg3, void *arg4, void *arg5, void *arg6, void *arg7)
 {
-	em_queued_call q = { function, 0 };
+	em_queued_call q = { function };
 	q.args[0].vp = arg1;
 	q.args[1].vp = arg2;
 	q.args[2].vp = arg3;
@@ -457,11 +555,73 @@ void EMSCRIPTEN_KEEPALIVE emscripten_main_thread_process_queued_calls()
 	// This must be before pthread_mutex_lock(), since pthread_mutex_lock() can call back to this function.
 	bool_inside_nested_process_queued_calls = 1;
 	pthread_mutex_lock(&call_queue_lock);
-	for (int i = 0; i < call_queue_length; ++i)
-		_do_call(call_queue[i]);
-	call_queue_length = 0;
-	bool_inside_nested_process_queued_calls = 0;
+	int head = emscripten_atomic_load_u32((void*)&call_queue_head);
+	int tail = emscripten_atomic_load_u32((void*)&call_queue_tail);
+	while (head != tail)
+	{
+		// Assume that the call is heavy, so unlock access to the call queue while it is being performed.
+		pthread_mutex_unlock(&call_queue_lock);
+		_do_call(call_queue[head]);
+		pthread_mutex_lock(&call_queue_lock);
+
+		head = (head + 1) % CALL_QUEUE_SIZE;
+		emscripten_atomic_store_u32((void*)&call_queue_head, head);
+		tail = emscripten_atomic_load_u32((void*)&call_queue_tail);
+	}
 	pthread_mutex_unlock(&call_queue_lock);
+
+	// If the queue was full and we had waiters pending to get to put data to queue, wake them up.
+	emscripten_futex_wake((void*)&call_queue_head, 0x7FFFFFFF);
+
+	bool_inside_nested_process_queued_calls = 0;
+}
+
+int emscripten_sync_run_in_main_runtime_thread_(EM_FUNC_SIGNATURE sig, void *func_ptr, ...)
+{
+	va_list args;
+	va_start(args, func_ptr);
+	int numArguments = EM_FUNC_SIG_NUM_FUNC_ARGUMENTS(sig);
+	em_queued_call q = { sig, func_ptr };
+	for(int i = 0; i < numArguments; ++i)
+		q.args[i].i = va_arg(args, int);
+	va_end(args);
+	emscripten_sync_run_in_main_thread(&q);
+	return q.returnValue.i;
+}
+
+void emscripten_async_run_in_main_runtime_thread_(EM_FUNC_SIGNATURE sig, void *func_ptr, ...)
+{
+	va_list args;
+	va_start(args, func_ptr);
+	int numArguments = EM_FUNC_SIG_NUM_FUNC_ARGUMENTS(sig);
+	em_queued_call *q = em_queued_call_malloc();
+	q->functionEnum = sig;
+	q->functionPtr = func_ptr;
+	for(int i = 0; i < numArguments; ++i)
+		q->args[i].i = va_arg(args, int);
+	va_end(args);
+	// 'async' runs are fire and forget, where the caller detaches itself from the call object after returning here,
+	// and it is the callee's responsibility to free up the memory after the call has been performed.
+	q->calleeDelete = 1;
+	emscripten_async_run_in_main_thread(q);
+}
+
+em_queued_call *emscripten_async_waitable_run_in_main_runtime_thread_(EM_FUNC_SIGNATURE sig, void *func_ptr, ...)
+{
+	va_list args;
+	va_start(args, func_ptr);
+	int numArguments = EM_FUNC_SIG_NUM_FUNC_ARGUMENTS(sig);
+	em_queued_call *q = em_queued_call_malloc();
+	q->functionEnum = sig;
+	q->functionPtr = func_ptr;
+	for(int i = 0; i < numArguments; ++i)
+		q->args[i].i = va_arg(args, int);
+	va_end(args);
+	// 'async waitable' runs are waited on by the caller, so the call object needs to remain alive for the caller to
+	// access it after the operation is done. The caller is responsible in cleaning up the object after done.
+	q->calleeDelete = 0;
+	emscripten_async_run_in_main_thread(q);
+	return q;
 }
 
 float EMSCRIPTEN_KEEPALIVE emscripten_atomic_load_f32(const void *addr)
@@ -727,3 +887,5 @@ uint64_t __atomic_fetch_xor_8(void *ptr, uint64_t value, int memmodel)
 {
 	return _emscripten_atomic_fetch_and_xor_u64(ptr, value);
 }
+
+weak_alias(__pthread_testcancel, pthread_testcancel);
